#[cfg(not(target_endian = "little"))]
compile_error!("Only little-endian architectures are supported.");

use crate::data::parser::{ParseError, Parser};

use std::cmp::Ordering;
use std::fmt::{self, Display, Formatter};
use std::io::{Read, Seek};

pub const DATA_LEN: usize = 1100 * 1200;

#[derive(Clone, Copy, Debug)]
pub struct VersionMismatch {
    pub actual: u8,
    pub expected: u8,
}

impl Display for VersionMismatch {
    fn fmt(&self, fmt: &mut Formatter<'_>) -> fmt::Result {
        fmt.write_fmt(format_args!(
            "Expected data version {}, got {}",
            self.expected, self.actual
        ))
    }
}

#[derive(Clone, Copy, Debug, Eq, Ord, PartialEq, PartialOrd)]
pub struct Datetime {
    pub year: u16,
    pub month: u8,
    pub day: u8,
    pub hour: u8,
    pub minute: u8,
}

#[derive(Debug)]
pub struct Rv {
    timestamp: Datetime,
    version: u8,
    precision: i8,
    interval: u16,
    forecast: u16,
    data: Vec<u16>,
}

impl Rv {
    pub fn parse<R: Read + Seek>(r: R) -> Result<Self, ParseError> {
        let mut parser = Parser::new(r);

        parser.expect("RV")?;

        let day = parser.parse_int::<_, 2>()?;
        let hour = parser.parse_int::<_, 2>()?;
        let minute = parser.parse_int::<_, 2>()?;

        let _wmo = parser.parse_int::<u32, 5>()?;

        let month = parser.parse_int::<_, 2>()?;
        let year = 2000 + parser.parse_int::<u16, 2>()?;

        parser.expect("BY")?;

        let _len = if parser.peek(7, 2)? == "VS" {
            parser.parse_int::<usize, 7>()?
        } else {
            parser.parse_int::<usize, 10>()?
        };

        parser.expect("VS")?;

        let version = parser.parse_int::<_, 2>()?;

        parser.expect("SW ")?;

        let _sw_version = parser.take(8)?;

        parser.expect("PR E")?;

        let precision = parser.parse_int::<_, 3>()?;

        parser.expect("INT")?;

        let interval = parser.parse_int::<_, 4>()?;

        parser.expect("GP")?;

        let _resolution = parser.take(9)?;

        parser.expect("VV ")?;

        let forecast = parser.parse_int::<_, 3>()?;

        parser.expect("MF ")?;

        let _flags = parser.parse_int::<u32, 8>()?;

        parser.expect("MS")?;

        let text_len = parser.parse_int::<_, 3>()?;

        let _text = parser.take(text_len)?;

        parser.expect("\x03")?;

        let data = parser.consume::<_, DATA_LEN>()?;

        Ok(Self {
            timestamp: Datetime {
                year,
                month,
                day,
                hour,
                minute,
            },
            version,
            precision,
            interval,
            forecast,
            data,
        })
    }

    pub fn expect_version(&self, version: u8) -> Result<(), VersionMismatch> {
        if self.version == version {
            Ok(())
        } else {
            Err(VersionMismatch {
                actual: self.version,
                expected: version,
            })
        }
    }

    pub fn scale(&self) -> f64 {
        10_f64.powi(self.precision as i32)
    }

    pub fn interval(&self) -> u16 {
        self.interval
    }

    pub fn instant(&self) -> (Datetime, u16) {
        (self.timestamp, self.forecast)
    }

    pub fn data(&self) -> &[u16] {
        &self.data[..]
    }
}

impl Eq for Rv {}

impl Ord for Rv {
    fn cmp(&self, other: &Self) -> Ordering {
        self.timestamp
            .cmp(&other.timestamp)
            .then(self.forecast.cmp(&other.forecast))
    }
}

impl PartialEq for Rv {
    fn eq(&self, other: &Self) -> bool {
        self.timestamp == other.timestamp && self.forecast == other.forecast
    }
}

impl PartialOrd for Rv {
    fn partial_cmp(&self, other: &Self) -> Option<Ordering> {
        Some(self.cmp(other))
    }
}
