mod config;
mod data;

use config::Config;

use data::{Rv, RV_DATA_LEN};

use image::imageops;
use image::RgbaImage;

use proj::Proj;

use std::cmp::Reverse;
use std::collections::BinaryHeap;
use std::env;
use std::fs::{self, File};
use std::io::BufReader;
use std::str;

const PROJ_WGS84_DE1200: &'static str = "+proj=stere +lat_0=90 +lat_ts=60 +lon_0=10 +a=6378137 +b=6356752.3142451802 +no_defs +x_0=543196.83521776402 +y_0=3622588.8619310018";

#[repr(u32)]
enum Intensity {
    None = 0xffffff00,
    Light = 0x80d0d080,
    Moderate = 0x50a0e080,
    Heavy = 0x1030f080,
    Unknown = 0xd0d0d080,
}

impl Intensity {
    fn from_rv_value(v: u16, scale: f64, interval: u16) -> Self {
        if v == 0x29c4 {
            Intensity::Unknown
        } else {
            let mm_per_h = v as f64 * scale * (60.0 / interval as f64);

            if mm_per_h < 0.1 {
                Intensity::None
            } else if mm_per_h < 5.0 {
                Intensity::Light
            } else if mm_per_h < 15.0 {
                Intensity::Moderate
            } else {
                Intensity::Heavy
            }
        }
    }
}

fn main() -> Result<(), String> {
    let config =
        Config::from_toml(&fs::read_to_string("config.toml").map_err(|err| err.to_string())?)
            .map_err(|err| err.to_string())?;

    let mut heap = BinaryHeap::new();

    for path in env::args().skip(1) {
        let rv = Rv::parse(BufReader::new(
            File::open(&path).map_err(|err| err.to_string())?,
        ))
        .map_err(|err| err.to_string())?;
        rv.expect_version(5).map_err(|err| err.to_string())?;

        heap.push(Reverse(rv));
    }

    let Reverse(current) = heap.pop().ok_or("At least one file must be given")?;

    let (timestamp, offset) = current.instant();

    if offset != 0 {
        eprintln!(
            "First file is expected to contain current situation, got forecast for +{} minutes",
            offset
        );
    }

    let mut img_data = Vec::with_capacity(RV_DATA_LEN * 4);

    for v in current.data() {
        let intensity = Intensity::from_rv_value(*v, current.scale(), current.interval());

        img_data.extend_from_slice(&(intensity as u32).to_be_bytes());
    }

    let mut img = RgbaImage::from_raw(1100, 1200, img_data).unwrap();
    imageops::flip_vertical_in_place(&mut img);
    img.save(format!(
        "rv_{}-{}-{}_{}:{}.png",
        timestamp.year, timestamp.month, timestamp.day, timestamp.hour, timestamp.minute
    ))
    .map_err(|err| err.to_string())?;

    let mut data = Vec::with_capacity(RV_DATA_LEN * heap.len());

    while heap.len() > 0 {
        let Reverse(forecast) = heap.pop().unwrap();
        data.extend_from_slice(forecast.data());
    }

    let wgs84_to_de1200 = Proj::new(PROJ_WGS84_DE1200).map_err(|err| err.to_string())?;

    for (name, poi) in config.pois() {
        let (x, y) = (wgs84_to_de1200
            .project(poi.to_radians(), false)
            .map_err(|err| err.to_string())?
            / 1000.0)
            .x_y();

        let (x, y) = (x.round() as usize, (1200 + y.round() as i64) as usize);
        if !(0..1100).contains(&x) || !(0..1200).contains(&y) {
            eprintln!("POI \"{}\" is out of range", name);
            continue;
        }

        let v = current.data()[y * 1100 + x];
        if v == 0x29c4 {
            eprintln!("No value for POI \"{}\"", name);
            continue;
        }

        let mm_per_h = v as f64 * current.scale() * (60.0 / current.interval() as f64);
        println!("intensity{{poi=\"{}\"}} {}", name, mm_per_h);
    }

    Ok(())
}
